from configparser import ConfigParser

from .base import FileHandlerBase


class IniHandler(FileHandlerBase):
    def load(self):
        cfg = ConfigParser()
        file_read = cfg.read(self._path)
        if not file_read:
            raise Exception("An error occurred while trying to read INI file '" +
                            self._path + "'")
        return {s: dict(cfg.items(s)) for s in cfg.sections()}

    def dump(self):
        cfg = ConfigParser()
        for key, value in self._config.items():
            if not isinstance(value, dict):
                raise AttributeError(
                    "INI configuration do not support nesting. section {} should contain dict".format(key))

        cfg.read_dict(self._config)
        with open(self._path, "w") as f:
            cfg.write(f)
        return True



import json

from .base import FileHandlerBase


class JsonHandler(FileHandlerBase):
    def load(self):
        with open(self._path, "r") as f:
            return json.load(f)

    def dump(self):
        with open(self._path, "w") as f:
            json.dump(self._config, f)
        return True



try:
    import yaml
except ImportError:
    raise ImportError("files.Yaml handler requires 'PyYAML' package\n"
                      "Install it with 'pip install pyyaml'")

from .base import FileHandlerBase


class YamlHandler(FileHandlerBase):
    def load(self):
        with open(self._path, "r") as f:
            return yaml.load(f)

    def dump(self):
        with open(self._path, "w") as f:
            yaml.dump(self._config, f)
        return True



class ConfigHandlerBase:

    def __init__(self):
        """
            Initialize the handler data store.
        """
        self._config = {}
        self._updated = False

    def load(self):
        """
            Load all configuration key values from the data store. Returns a nested dict.
            :rtype: dict
        """
        return {}

    def dump(self):
        """
            Serialize and store the configuration key, values to the data store.
            rest self._updated flag to False
            self._updated = False
            :rtype: bool (success)
        """
        raise NotImplementedError

    def flush(self):
        """
            flush in memory configuration.
            :rtype: bool (success)
        """
        self._config = {}
        self._updated = True
        return True

    def destroy(self):
        """
            destroys persistent configuration. configuration is removed from disk or database
            :rtype: bool (success)
        """
        raise NotImplementedError

    def as_dict(self):
        """
            Serialize the current configuration key, values.
            :rtype: dict
        """
        if self._updated:
            # Listen to yourself
            # Always read from the data store
            self.dump()
            self._config = self.load()
        return self._config



import os

from config42.handlers import ConfigHandlerBase


class FileHandlerBase(ConfigHandlerBase):
    def __init__(self, *, path):
        super().__init__()
        self._path = path
        try:
            config = self.load()
        except Exception:
            config = None
        self._config = config if config is not None else {}

    def destroy(self):
        """
            remove persistent configuration file.
            :rtype: bool (success)
        """
        os.remove(self._path)
        return True



from config42.handlers import ConfigHandlerBase

try:
    from etcd import Client, EtcdKeyNotFound
except ImportError:
    raise ImportError("Please install python-etcd package")


class Etcd(ConfigHandlerBase):
    def __init__(self, keyspace=None, **kwargs):
        """
            Initialize the handler data store.
            :param keyspace: etcd keyspace for configuration map starting with /
            :type key: string
            :param kwargs: generic params forwarded from the Configmanager
            :type key: dict
        """
        super().__init__()
        self.client = Client(**kwargs)
        self.keyspace = keyspace if keyspace else '/config'
        self._config = self.load()

    def load(self):
        """
            Load all configuration key values from the Etcd data store. Returns a nested dict.
            :rtype: dict
        """
        try:
            directory = self.client.read(self.keyspace, recursive=True)
            return self.recursive(directory, prefix=self.keyspace)
        except EtcdKeyNotFound:
            return {}

    def dump(self):
        """
            Serialize and store the configuration key, values to the Etcd data store.
            :rtype: bool (success)
        """
        flat_dict = self.flatten(self._config, parent_key=self.keyspace, seperator='/')
        for key, value in flat_dict.items():
            self.client.set(key, value)
        self._updated = False
        return True

    def destroy(self):
        """
            destroys given keyspace in etcd
            :rtype: bool (success)
        """
        try:
            self.client.delete(self.keyspace, recursive=True, dir=True)
            return True
        except KeyError:
            return False

    def recursive(self, directory, prefix='', seperator='/'):
        items = {}
        for item in directory.leaves:
            if item.dir:
                items.update(self.recursive(item), prefix=item.key, seperator=seperator)
            else:
                items[item.key.replace(prefix, '').replace(seperator, '')] = item.value
        return dict(items)

    def flatten(self, dict_, parent_key='', seperator='/'):
        items = []
        for key, value in dict_.items():
            new_key = (parent_key + seperator + key) if parent_key else key
            if isinstance(value, dict):
                items.extend(self.flatten(value, parent_key=new_key, seperator=seperator).items())
            else:
                items.append((new_key, value))
        return dict(items)



from config42.handlers import ConfigHandlerBase


class Memory(ConfigHandlerBase):
    def __init__(self):
        """
            Initializes the handler data store.
            :param kwargs: generic params forwarded from the Configmanager
            :type key: dict
        """
        super().__init__()
        self.in_memory_config = {}
        self._config = {}

    def load(self):
        """
            As it is stored in memory, returns the last committed dict.
            :rtype: dict
        """
        return self.in_memory_config

    def dump(self):
        """
            Affects current configuration .
            :rtype: bool (success)
        """
        self.in_memory_config = self._config
        self._updated = False
        return True

    def destroy(self):
        """
            Destroys the committed configuration
            :rtype: bool (success)
        """
        self.in_memory_config = {}
        return True



import os

from setuptools import setup, find_packages

try:
    here = os.path.abspath(os.path.dirname(__file__))
    README = open(os.path.join(here, "README.md")).read()
    with open(os.path.join(here, "requirements/base.txt")) as f:
        required = [l.strip('\n') for l in f if
                    l.strip('\n') and not l.startswith('#')]
except IOError:
    required = []
    README = ""

setup(
    name="config42",
    packages=find_packages(),
    version='0.2.3',
    license='GPLv3+',
    description="Configuration manager for cloud native application, support configuration stored"
                " in memory, in files, in databases",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Amine Ben Asker",
    author_email="ben.asker.amine@gmail.com",
    url="https://github.com/yurilaaziz/config42",
    download_url='https://github.com/yurilaaziz/config42/releases/tag/0.1',
    keywords="Pretty configuration manager, Key value data store, cloud native configuration",
    install_requires=required,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



from config42.handlers import Memory


class ConfigManager:
    def __init__(self, defaults=None, handler=None, **handler_kwargs):
        self.defaults = defaults if defaults else {}

        if not handler:
            handler = Memory

        self.handler = handler(**handler_kwargs)

    def get_defaults(self, key):
        return self.recursive(key, obj=self.defaults)

    def recursive(self, key, obj, value=None):
        _keys = key.split('.')
        assert len(_keys) > 0

        if isinstance(obj, list):
            _base = int(_keys[0])
        else:
            _base = _keys[0]

        if isinstance(obj, dict) and _base not in obj and not value:
            return None
        if isinstance(obj, list) and _base >= len(obj) and not value:
            return None
        elif isinstance(obj, list) and value:
            raise AttributeError("Insertion in list is not allowed")

        if len(_keys) == 1:
            if value:
                obj[_base] = value
            return obj[_base]
        else:
            if isinstance(obj, dict) and _base not in obj:
                obj[_base] = dict()

            return self.recursive('.'.join(_keys[1:]), obj[_base], value)

    def get(self, key):
        """
            Gets config value for a given key from the data store. Returns the value that matches the supplied key.
            If the value is not set, a default value will be returned as set by set_defaults.
            :param key: The configuration key to return a config value for
            :type key: str
            :rtype: Any supported (str, int, bool, list-of-supported-types)
        """
        value = self.recursive(key, obj=self.handler.as_dict())
        return value if value is not None else self.get_defaults(key)

    def set(self, key, value, trigger_commit=True):
        """
            Sets config value for a given key in the data store.
            :param key: The configuration key to set
            :type key: str
            :param value: The value to set the configuration key to
            :type value: Any supported (str, int, bool, list-of-supported-types)
            :param trigger_commit: Flag whether to trigger a config commit after all values are set.
            :type trigger_commit: bool
            :rtype: bool (success)
        """
        if self.recursive(key, obj=self.handler._config) == value:
            return False  # Not updating

        self.recursive(key, obj=self.handler._config, value=value)
        self.handler._updated = True
        if trigger_commit:
            self.commit()

        return self.handler._updated

    def replace(self, config, trigger_commit=True):
        """
        Completely reset the config with a set of key values.
        It is the equivalent of unsetting all keys, followed by a
        set_many. Anything not in the supplied config will revert to default.
        :param config: A dictionary of keys and values to set as defaults
        :type config: dict
        :param trigger_commit: Flag whether to trigger a config commit after all values are set.
        :type trigger_commit: bool
        """
        self.handler.flush()
        self.set_many(config, trigger_commit)

    def set_many(self, config, trigger_commit=True):
        """
        Sets the value of multiple config settings simultaneously.
        This postpones the
        triggering of the commit signal after all values are set.
        :param config: A dictionary of keys and values to set.
        :type key: dict
        :param trigger_commit: Flag whether to trigger a config commit after all values are set.
        :type trigger_commit: bool
        """
        has_been_updated = False
        for key, value in config.items():
            updated = self.set(key, value, trigger_commit=False)
            has_been_updated = has_been_updated or updated

        if has_been_updated and trigger_commit:
            self.commit()

    def commit(self):
        return self.handler.dump()



from .ini import IniHandler
from .json import JsonHandler
from .yaml import YamlHandler


class FileHandler(object):
    def __new__(cls, *, path, extension=None, **kwargs):
        """
            :param path: path of the config file
        """
        handler_map = {
            'yaml': YamlHandler,
            'yml': YamlHandler,
            'json': JsonHandler,
            'ini': IniHandler
        }
        path = path.lower()
        if not extension:
            extension = path.split('.')[-1]

        if handler_map.get(extension):
            return handler_map.get(extension)(path=path, **kwargs)
        else:
            raise NotImplementedError("Only {} extensions are supported".format(", ".join(handler_map.keys())))
